"""
Implementation of the Yubico OTP algorithm. This can generate and parse OTP
structures.

>>> from binascii import unhexlify
>>> key = b'0123456789abcdef'
>>> otp = OTP(unhexlify(b'0123456789ab'), 5, 0x0153f8, 0, 0x1234)
>>> _ = repr(otp)  # coverage
>>> _ = str(otp)  # coverage
>>> token = encode_otp(otp, key, b'cclngiuv')
>>> token == b'cclngiuvttkhthcilurtkerbjnnkljfkjccklkhl'
True
>>> public_id, otp2 = decode_otp(token, key)
>>> public_id == b'cclngiuv'
True
>>> otp2 == otp
True
"""
from binascii import hexlify
from datetime import datetime
from random import randrange
from struct import pack, unpack

from Crypto.Cipher import AES

from .crc import crc16, verify_crc16
from .modhex import is_modhex, modhex, unmodhex


__all__ = ['decode_otp', 'encode_otp', 'OTP', 'YubiKey', 'CRCError']


class CRCError(ValueError):
    """
    Raised when a decrypted token has an invalid checksum.
    """
    pass


def decode_otp(token, key):
    """
    Decodes a modhex-encoded Yubico OTP token and returns the public ID and the
    unpacked :class:`OTP` object.

    :param bytes token: A modhex-encoded buffer, as generated by a YubiKey
        device. Decoded, this should consist of 0-16 bytes of public ID
        followed by 16 bytes of encrypted OTP data.
    :param bytes key: A 16-byte AES key as a binary string.

    :returns: The public ID in its modhex-encoded form and the OTP structure.
    :rtype: (bytes, :class:`OTP`)

    :raises: ``ValueError`` if the string can not be decoded.
    :raises: :exc:`CRCError` if the checksum on the decrypted data is
        incorrect.
    """
    if len(key) != 16:
        raise ValueError('Key must be exactly 16 bytes')

    public_id, token = token[:-32], token[-32:]

    buf = unmodhex(token)
    buf = AES.new(key, AES.MODE_ECB).decrypt(buf)
    otp = OTP.unpack(buf)

    return (public_id, otp)


def encode_otp(otp, key, public_id=b''):
    """
    Encodes an :class:`OTP` structure, encrypts it with the given key and
    returns the modhex-encoded token.

    :param otp: The OTP structure.
    :type otp: :class:`OTP`

    :param bytes key: A 16-byte AES key as a binary string.
    :param bytes public_id: An optional public id, modhex-encoded. This can be
        at most 32 bytes.

    :raises: ValueError if any parameters are out of range.
    """
    if len(key) != 16:
        raise ValueError('Key must be exactly 16 bytes')

    if not is_modhex(public_id):
        raise ValueError('public_id must be a valid modhex string')

    if len(public_id) > 32:
        raise ValueError('public_id may be no longer than 32 modhex characters')

    buf = otp.pack()
    buf = AES.new(key, AES.MODE_ECB).encrypt(buf)
    token = modhex(buf)

    return public_id + token


class OTP(object):
    """
    A single YubiKey OTP. This is typically instantiated by parsing an encoded
    OTP.

    :param bytes uid: The private ID as a 6-byte binary string.
    :param int session: The non-volatile usage counter.
    :param int timestamp: An integer in [0..2^24].
    :param int counter: The volatile usage counter.
    :param int rand: An arbitrary number in [0..2^16].
    """
    def __init__(self, uid, session, timestamp, counter, rand):
        self.uid = uid
        self.session = session
        self.timestamp = timestamp
        self.counter = counter
        self.rand = rand

    def __repr__(self):
        return 'OTP({self.uid!r}, {self.session!r}, {self.timestamp!r}, {self.counter!r}, {self.rand!r})'.format(self=self)

    def __str__(self):
        return 'OTP: {0} {1}/{2} (0x{3:x}/0x{4:x})'.format(hexlify(self.uid), self.session, self.counter, self.timestamp, self.rand)

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False

        self_props = (self.uid, self.session, self.timestamp, self.counter, self.rand)
        other_props = (other.uid, other.session, other.timestamp, other.counter, other.rand)

        return (self_props == other_props)

    def pack(self):
        """
        Returns the OTP packed into a binary string, ready to be encrypted and
        encoded.
        """
        fields = (
            self.uid,
            self.session,
            self.timestamp & 0xff, (self.timestamp >> 8) & 0xffff,
            self.counter,
            self.rand,
        )

        buf = pack('<6s H BH B H', *fields)

        crc = ~crc16(buf) & 0xffff
        buf += pack('<H', crc)

        return buf

    @classmethod
    def unpack(cls, buf):
        """
        Parse a packed OTP. This is the complement to :meth:`pack` so the
        buffer should be a decoded, decrypted OTP buffer.

        :param bytes buf: A packed OTP structure.
        :raises: :exc:`CRCError` if the buffer does not pass crc validation.
        """
        if not verify_crc16(buf):
            raise CRCError('OTP checksum is invalid')

        uid, session, t1, t2, counter, rand, crc = unpack('<6s H BH B H H', buf)

        timestamp = (t2 << 8) | t1

        return cls(uid, session, timestamp, counter, rand)


class YubiKey(object):
    """
    A simulated YubiKey device. This can be used to generate a sequence of
    Yubico OTP tokens.

    :param bytes uid: The private ID as a 6-byte binary string.
    :param int session: The non-volatile usage counter. It is the caller's
        responsibility to persist this. Note that this may increment if the
        volatile counter wraps, so you should only increment and persist this
        after you have finished generating tokens.
    :param int counter: The volatile session counter. This defaults to 0 at
        init time, but the caller can override this.
    """
    def __init__(self, uid, session, counter=0):
        self.uid = uid
        self.session = min(session, 0x7fff)
        self.counter = min(counter, 0xff)

        self._init_timestamp()

    def generate(self):
        """
        Return a new OTP object, as if the user had pressed the YubiKey button.

        :rtype: :class:`OTP`
        """
        otp = OTP(self.uid, self.session, self._timestamp(), self.counter, randrange(0xffff))
        self._increment_counter()

        return otp

    def _init_timestamp(self):
        self._timestamp_base = randrange(0x00ffff)
        self._timestamp_start = datetime.now()

    def _timestamp(self):
        """
        Returns the current timestamp value, based on the number of seconds
        since the object was created.
        """
        delta = datetime.now() - self._timestamp_start
        delta = delta.days * 86400 + delta.seconds

        return (self._timestamp_base + (delta * 8)) % 0xffffff

    def _increment_counter(self):
        if self.counter >= 0xff:
            self._increment_session()
            self.counter = 0
        else:
            self.counter += 1

    def _increment_session(self):
        self.session = min(self.session + 1, 0x7fff)



"""
CRC16 implementation for Yubico OTP.
"""


def crc16(data):
    """
    Generate the crc-16 value for a byte string.

    >>> from binascii import unhexlify
    >>> c = crc16(unhexlify(b'8792ebfe26cc130030c20011c89f'))
    >>> hex(~c & 0xffff)
    '0xc823'
    >>> v = crc16(unhexlify(b'8792ebfe26cc130030c20011c89f23c8'))
    >>> hex(v)
    '0xf0b8'
    """
    crc = 0xffff

    for byte in iter(data):
        crc ^= byte

        for i in range(8):
            lsb = crc & 1
            crc >>= 1
            if lsb == 1:
                crc ^= 0x8408

    return crc


def verify_crc16(data):
    """
    Return true if this given byte string has a valid crc-16 residual.

    >>> from binascii import unhexlify
    >>> verify_crc16(unhexlify(b'8792ebfe26cc130030c20011c89f23c8'))
    True
    >>> verify_crc16(unhexlify(b'0792ebfe26cc130030c20011c89f23c8'))
    False
    """
    return crc16(data) == 0xf0b8



from doctest import DocTestSuite
import unittest

from . import crc, modhex, otp


def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()

    suite.addTests(tests)
    suite.addTest(DocTestSuite(crc))
    suite.addTest(DocTestSuite(modhex))
    suite.addTest(DocTestSuite(otp))

    return suite



#!/usr/bin/env python

from setuptools import setup


setup(
    name='YubiOTP',
    version='1.0.0.post1',
    description='A library for verifying YubiKey OTP tokens, both locally and through a Yubico web service.',
    author='Peter Sagerson',
    author_email='psagers@ignorare.net',
    url='https://github.com/django-otp/yubiotp',
    project_urls={
        "Documentation": 'https://yubiotp.readthedocs.io/',
        "Source": 'https://github.com/django-otp/yubiotp',
    },
    license='BSD',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Security',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],

    packages=[
        'yubiotp',
    ],
    scripts=[
        'bin/yubikey',
        'bin/yubiclient',
    ],
    install_requires=[
        'pycryptodome',
    ],
)



"""
Implementation of `modhex encoding <http://www.yubico.com/modhex-calculator>`_,
which uses keyboard-independent characters.

::

    hex digit:    0123456789abcdef
    modhex digit: cbdefghijklnrtuv
"""

from binascii import hexlify, unhexlify
from functools import partial
import struct


__all__ = ['modhex', 'unmodhex', 'is_modhex', 'hex_to_modhex', 'modhex_to_hex']


def modhex(data):
    """
    Encode a string of bytes as modhex.

    >>> modhex(b'abcdefghijklmnop') == b'hbhdhehfhghhhihjhkhlhnhrhthuhvic'
    True
    """
    return hex_to_modhex(hexlify(data))


def unmodhex(encoded):
    """
    Decode a modhex string to its binary form.

    >>> unmodhex(b'hbhdhehfhghhhihjhkhlhnhrhthuhvic') == b'abcdefghijklmnop'
    True
    """
    return unhexlify(modhex_to_hex(encoded))


def is_modhex(encoded):
    """
    Returns ``True`` iff the given string is valid modhex.

    >>> is_modhex(b'cbdefghijklnrtuv')
    True
    >>> is_modhex(b'cbdefghijklnrtuvv')
    False
    >>> is_modhex(b'cbdefghijklnrtuvyy')
    False
    """
    if any(c not in modhex_chars for c in encoded):
        return False
    elif len(encoded) % 2 != 0:
        return False
    else:
        return True


def hex_to_modhex(hex_str):
    """
    Convert a string of hex digits to a string of modhex digits.

    >>> hex_to_modhex(b'69b6481c8baba2b60e8f22179b58cd56') == b'hknhfjbrjnlnldnhcujvddbikngjrtgh'
    True
    >>> hex_to_modhex(b'6j')
    Traceback (most recent call last):
        ...
    ValueError: Illegal hex character in input
    """
    try:
        return b''.join(int2byte(hex_to_modhex_char(b))
                        for b in iter(hex_str.lower()))
    except ValueError:
        raise ValueError('Illegal hex character in input')


def modhex_to_hex(modhex_str):
    """
    Convert a string of modhex digits to a string of hex digits.

    >>> modhex_to_hex(b'hknhfjbrjnlnldnhcujvddbikngjrtgh') == b'69b6481c8baba2b60e8f22179b58cd56'
    True
    >>> modhex_to_hex(b'hbhdxx')
    Traceback (most recent call last):
        ...
    ValueError: Illegal modhex character in input
    """
    try:
        return b''.join(int2byte(modhex_to_hex_char(b))
                        for b in iter(modhex_str.lower()))
    except ValueError:
        raise ValueError('Illegal modhex character in input')


#
# Internals
#

def int2byte(i):
    return struct.Struct(">B").pack(i)


def lookup(alist, key):
    try:
        return next(v for (k, v) in alist if k == key)
    except StopIteration:
        raise ValueError()


hex_chars = b'0123456789abcdef'
modhex_chars = b'cbdefghijklnrtuv'

hex_to_modhex_map = list(zip(iter(hex_chars), iter(modhex_chars)))
modhex_to_hex_map = list(zip(iter(modhex_chars), iter(hex_chars)))

hex_to_modhex_char = partial(lookup, hex_to_modhex_map)
modhex_to_hex_char = partial(lookup, modhex_to_hex_map)
